# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
This module contains functions for matching coordinate catalogs.
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import numpy as np

from astropy.extern import six
from astropy.coordinates.representation import UnitSphericalRepresentation

__all__ = ['match_coordinates_3d', 'match_coordinates_sky']


def match_coordinates_3d(matchcoord, catalogcoord, nthneighbor=1, storekdtree='_kdtree_3d'):
    """
    Finds the nearest 3-dimensional matches of a coordinate or coordinates in
    a set of catalog coordinates.

    This finds the 3-dimensional closest neighbor, which is only different
    from the on-sky distance if ``distance`` is set in either ``matchcoord``
    or ``catalogcoord``.

    Parameters
    ----------
    matchcoord : `~astropy.coordinates.BaseCoordinateFrame` or `~astropy.coordinates.SkyCoord`
        The coordinate(s) to match to the catalog.
    catalogcoord : `~astropy.coordinates.BaseCoordinateFrame` or `~astropy.coordinates.SkyCoord`
        The base catalog in which to search for matches. Typically this will
        be a coordinate object that is an array (i.e.,
        ``catalogcoord.isscalar == False``)
    nthneighbor : int, optional
        Which closest neighbor to search for.  Typically ``1`` is desired here,
        as that is correct for matching one set of coordinates to another.
        The next likely use case is ``2``, for matching a coordinate catalog
        against *itself* (``1`` is inappropriate because each point will find
        itself as the closest match).
    storekdtree : bool or str, optional
        If a string, will store the KD-Tree used for the computation
        in the ``catalogcoord``, as an attribute in ``catalogcoord`` with the
        provided name.  This dramatically speeds up subsequent calls with the
        same catalog. If False, the KD-Tree is discarded after use.

    Returns
    -------
    idx : integer array
        Indecies into ``catalogcoord`` to get the matched points for each
        ``matchcoord``. Shape matches ``matchcoord``.
    sep2d : `~astropy.coordinates.Angle`
        The on-sky separation between the closest match for each ``matchcoord``
        and the ``matchcoord``. Shape matches ``matchcoord``.
    dist3d : `~astropy.units.Quantity`
        The 3D distance between the closest match for each ``matchcoord`` and
        the ``matchcoord``. Shape matches ``matchcoord``.

    Notes
    -----
    This function requires `SciPy <http://www.scipy.org>`_ to be installed
    or it will fail.
    """
    from warnings import warn

    #without scipy this will immediately fail
    from ._kdtree import KDTree

    if storekdtree is True:  # backwards compatibility for pre v0.4
        storekdtree = '_kdtree'

    # figure out where any cached KDTree might be
    if isinstance(storekdtree, six.string_types):
        kdt = getattr(catalogcoord, storekdtree, None)
        if kdt is not None and not isinstance(kdt, KDTree):
            raise ValueError('Invalid `storekdtree` string:' + storekdtree)
    elif isinstance(storekdtree, KDTree):
        kdt = storekdtree
        storekdtree = None
    elif not storekdtree:
        kdt = None
    else:
        raise ValueError('Invalid `storekdtree` argument:' +
                          str(storekdtree))

    if kdt is None:
        #need to build the cartesian KD-tree for the catalog
        cartxyz = catalogcoord.cartesian.xyz
        flatxyz = cartxyz.reshape((3, np.prod(cartxyz.shape) // 3))
        kdt = KDTree(flatxyz.value.T)

    #make sure coordinate systems match
    matchcoord = matchcoord.transform_to(catalogcoord)

    #make sure units match
    catunit = catalogcoord.cartesian.x.unit
    matchxyz = matchcoord.cartesian.xyz.to(catunit)

    matchflatxyz = matchxyz.reshape((3, np.prod(matchxyz.shape) // 3))
    dist, idx = kdt.query(matchflatxyz.T, nthneighbor)

    if nthneighbor > 1:  # query gives 1D arrays if k=1, 2D arrays otherwise
        dist = dist[:, -1]
        idx = idx[:, -1]

    if storekdtree:
        #cache the kdtree in `catalogcoord`
        setattr(catalogcoord, storekdtree, kdt)

    sep2d = catalogcoord[idx].separation(matchcoord)
    return idx.reshape(matchxyz.shape[1:]), sep2d, dist.reshape(matchxyz.shape[1:]) * catunit


def match_coordinates_sky(matchcoord, catalogcoord, nthneighbor=1, storekdtree='_kdtree_sky'):
    """
    Finds the nearest on-sky matches of a coordinate or coordinates in
    a set of catalog coordinates.

    This finds the on-sky closest neighbor, which is only different from the
    3-dimensional match if ``distance`` is set in either ``matchcoord``
    or ``catalogcoord``.

    Parameters
    ----------
    matchcoord : `~astropy.coordinates.BaseCoordinateFrame` or `~astropy.coordinates.SkyCoord`
        The coordinate(s) to match to the catalog.
    catalogcoord : `~astropy.coordinates.BaseCoordinateFrame` or `~astropy.coordinates.SkyCoord`
        The base catalog in which to search for matches. Typically this will
        be a coordinate object that is an array (i.e.,
        ``catalogcoord.isscalar == False``)
    nthneighbor : int, optional
        Which closest neighbor to search for.  Typically ``1`` is desired here,
        as that is correct for matching one set of coordinates to another.
        The next likely use case is ``2``, for matching a coordinate catalog
        against *itself* (``1`` is inappropriate because each point will find
        itself as the closest match).
    storekdtree : bool or str, optional
        If a string, will store the KD-Tree used for the computation
        in the ``catalogcoord``, as an attrbute in ``catalogcoord`` with the
        provided name.  This dramatically speeds up subsequent calls with the
        same catalog. If False, the KD-Tree is discarded after use.

    Returns
    -------
    idx : integer array
        Indecies into ``catalogcoord`` to get the matched points for each
        ``matchcoord``. Shape matches ``matchcoord``.
    sep2d : `~astropy.coordinates.Angle`
        The on-sky separation between the closest match for each
        ``matchcoord`` and the ``matchcoord``. Shape matches ``matchcoord``.
    dist3d : `~astropy.units.Quantity`
        The 3D distance between the closest match for each ``matchcoord`` and
        the ``matchcoord``. Shape matches ``matchcoord``.  If either
        ``matchcoord`` or ``catalogcoord`` don't have a distance, this is the 3D
        distance on the unit sphere, rather than a true distance.

    Notes
    -----
    This function requires `SciPy <http://www.scipy.org>`_ to be installed
    or it will fail.
    """

    # send to catalog frame
    newmatch = matchcoord.transform_to(catalogcoord)

    #strip out distance info
    match_urepr = newmatch.data.represent_as(UnitSphericalRepresentation)
    newmatch_u = newmatch.realize_frame(match_urepr)

    cat_urepr = catalogcoord.data.represent_as(UnitSphericalRepresentation)
    newcat_u = catalogcoord.realize_frame(cat_urepr)

    idx, sep2d, sep3d = match_coordinates_3d(newmatch_u, newcat_u, nthneighbor, storekdtree)
    # sep3d is *wrong* above, because the distance information was removed,
    # unless one of the catalogs doesn't have a real distance
    if not (isinstance(catalogcoord.data, UnitSphericalRepresentation) or
            isinstance(newmatch.data, UnitSphericalRepresentation)):
        sep3d = catalogcoord[idx].separation_3d(newmatch)

    #update the kdtree on the actual passed-in coordinate
    if storekdtree:
        setattr(catalogcoord, storekdtree, getattr(newcat_u, storekdtree))

    return idx, sep2d, sep3d



#! /usr/bin/env python
# Runs an example of each operation
import lsmtool
import os


s = lsmtool.load('tests/no_patches.sky')
s2 = lsmtool.load('tests/apparent.sky')


def test_select():
    print('Select individual sources with Stokes I fluxes above 1 Jy')
    s.select('I > 1.0 Jy')
    assert len(s) == 965


def test_transfer():
    print('Transfer patches from patches.sky')
    s.transfer('tests/patches.sky')
    assert s.hasPatches


def test_remove():
    print('Remove patches with total fluxes below 2 Jy')
    s.remove('I < 2.0 Jy', aggregate='sum')
    assert len(s) == 433


def test_upgroup():
    print('Ungroup the skymodel')
    s.ungroup()
    assert ~s.hasPatches


def test_concatenate():
    print('Concatenate with concat.sky')
    s.concatenate('tests/concat.sky', matchBy = 'position', radius = '30 arcsec', keep = 'from2')
    assert len(s) == 2165


def test_concatenate_differing_spectral_index():
    print('Concatenate with single_spectralindx.sky')
    s4 = lsmtool.load('tests/no_patches.sky')
    s4.concatenate('tests/single_spectralindx.sky', matchBy = 'position', radius = '30 arcsec', keep = 'from2')
    assert len(s) == 2165


def test_compare():
    print('Compare to concat.sky')
    if os.path.exists('tests/flux_ratio_vs_distance.sky'):
        os.remove('tests/flux_ratio_vs_distance.sky')
    c = lsmtool.load('tests/concat.sky')
    c.ungroup()
    c.select('I > 5.0 Jy')
    s.ungroup()
    s.compare(c, outDir='tests/')
    assert os.path.exists('tests/flux_ratio_vs_distance.pdf')


def test_add():
    print('Add a source')
    s.add({'Name': 'src1', 'Type': 'POINT', 'Ra': 277.4232, 'Dec': 48.3689, 'I': 0.69})
    assert len(s) == 2166


def test_group():
    print('Group using tessellation to a target flux of 50 Jy')
    s.group('tessellate', targetFlux = '50.0 Jy')
    assert s.hasPatches


def test_move():
    print('Move patch Patch_1 to 16:04:16.2288, 58.03.06.912')
    s.move('Patch_1', position=['16:04:16.2288', '58.03.06.912'])
    assert round(s.getPatchPositions()['Patch_1'][0].value, 4) == 241.0676


def test_merge():
    print('Merge patches Patch_0 and Patch_2')
    s.merge(['Patch_0', 'Patch_2'], name = 'merged_patch')
    assert len(s.getPatchNames()) == 139


def test_setPatchPositions():
    print('Set patch positions to midpoint of patch')
    s.setPatchPositions(method='mid')
    assert round(s.getPatchPositions()['merged_patch'][0].value, 4) == 274.2124


def test_write():
    print('Write final model to file')
    if os.path.exists('tests/final.sky'):
        os.remove('tests/final.sky')
    s.write('tests/final.sky', clobber=True)
    assert os.path.exists('tests/final.sky')


def test_plot():
    print('Plot the model')
    if os.path.exists('tests/plot.pdf'):
        os.remove('tests/plot.pdf')
    s.plot('tests/plot.pdf')
    assert os.path.exists('tests/plot.pdf')


def test_meanshift():
    print('Group the model with the meanshift algorithm')
    s2.group('meanshift', byPatch=True, lookDistance=0.075, groupingDistance=0.01)
    assert len(s2.getPatchPositions()) == 67



# -*- coding: utf-8 -*-
#
# Defines the custom logger

import logging

# set the logging format and default level (info)
logging.basicConfig(format='%(levelname)s: %(message)s', level=logging.INFO)

def setLevel(level):
    """
    Change the verbosity of the logger

    Parameters
    ----------
    level : str
        Desired level (from less to more verbosity): 'warning', 'info', 'debug'
    """
    level = level.lower()
    if level not in ['warning', 'info', 'debug']:
        raise ValueError("Invalid level name specified. Level name must be one "
            "of: 'warning', 'info', 'debug'")
    if level == 'warning':
        logging.root.setLevel(logging.WARNING)
    elif level == 'info':
        logging.root.setLevel(logging.INFO)
    elif level == 'debug':
        logging.root.setLevel(logging.DEBUG)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# This operation implements removal of sources using a filter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import logging
log = logging.getLogger('LSMTool.REMOVE')
log.debug('Loading REMOVE module.')


def run(step, parset, LSM):

    outFile = parset.getString('.'.join(["LSMTool.Steps", step, "OutFile"]), '' )
    filterExpression = parset.getString('.'.join(["LSMTool.Steps", step, "FilterExpression"]), '' )
    aggregate = parset.getString('.'.join(["LSMTool.Steps", step, "Aggregate"]), '' )
    applyBeam = parset.getBool('.'.join(["LSMTool.Steps", step, "ApplyBeam"]), False )

    if filterExpression == '':
        filterExpression = None
    if aggregate == '':
        aggregate = None

    try:
        remove(LSM, filterExpression, aggregate=aggregate, applyBeam=applyBeam)
        result = 0
    except Exception as e:
        log.error(e)
        result = 1

    # Write to outFile
    if outFile != '' and result == 0:
        LSM.write(outFile, clobber=True)

    return result


def remove(LSM, filterExpression, aggregate=None, applyBeam=None,
    useRegEx=False, force=True):
    """
    Filters the sky model, removing all sources that meet the given expression.

    After filtering, the sky model contains only those sources for which the
    given filter expression is false.

    Parameters
    ----------
    LSM : SkyModel object
        Input sky model
    filterExpression : str, dict, list, or numpy array
        - If string:
            A string specifying the filter expression in the form:
            '<property> <operator> <value> [<units>]'
            (e.g., 'I <= 10.5 Jy')

        - If dict:
            The filter can also be given as a dictionary in the form:
            {'filterProp':property, 'filterOper':operator,
                'filterVal':value, 'filterUnits':units}

        - If list:
            The filter can also be given as a list of:
            [property, operator, value] or
            [property, operator, value, units]

        - If numpy array:
            The indices to filter on can be specified directly as a numpy array
            of row or patch indices such as:
            array([ 0,  2, 19, 20, 31, 37])

            or as a numpy array of bools with the same length as the sky model.

            If a numpy array is given and the indices correspond to patches, then
            set aggregate=True

        The property to filter on must be one of the following:
            - a valid column name
            - the filename of a mask image

        Supported operators are:
            - !=
            - <=
            - >=
            - >
            - <
            - = (or '==')
        Units are optional and must be specified as required by astropy.units
    aggregate : str, optional
        If set, the array returned will be of values aggregated
        over the patch members. The following aggregation functions are
        available:
            - 'sum': sum of patch values
            - 'mean': mean of patch values
            - 'wmean': Stokes I weighted mean of patch values
            - 'min': minimum of patch values
            - 'max': maximum of patch values
            - True: only valid when the filter indices are specified directly as
                a numpy array. If True, filtering is done on patches instead of
                sources
    applyBeam : bool, optional
        If True, apparent fluxes will be used
    useRegEx : bool, optional
        If True, string matching will use regular expression matching. If
        False, string matching uses Unix filename matching

    Examples
    --------
    Filter on column 'I' (Stokes I flux). This filter will remove all sources
    with Stokes I flux greater than 1.5 Jy::

        >>> LSM = lsmtool.load('sky.model')
        >>> remove(LSM, 'I > 1.5 Jy')
        INFO: Removed 1102 sources.

    If the sky model has patches and the filter is desired per patch, use
    ``aggregate = function``. For example, to select on the sum of the patch
    fluxes::

        >>> remove(LSM, 'I > 1.5 Jy', aggregate='sum')

    Filter on source names, removing those that match "src*_1?"::

        >>> remove(LSM, 'Name == src*_1?')

    Use a CASA clean mask image to remove sources that lie in masked regions::

        >>> remove(LSM, 'clean_mask.mask == True')

    """
    from . import _filter

    _filter.filter(LSM, filterExpression, aggregate=aggregate,
        applyBeam=applyBeam, useRegEx=useRegEx, exclusive=True, force=force)



# Copyright Anne M. Archibald 2008
# Released under the scipy license
import sys
import numpy as np
from heapq import heappush, heappop
import scipy.sparse

def minkowski_distance_p(x,y,p=2):
    """Compute the pth power of the L**p distance between x and y

    For efficiency, this function computes the L**p distance but does
    not extract the pth root. If p is 1 or infinity, this is equal to
    the actual L**p distance.
    """
    x = np.asarray(x)
    y = np.asarray(y)
    if p==np.inf:
        return np.amax(np.abs(y-x),axis=-1)
    elif p==1:
        return np.sum(np.abs(y-x),axis=-1)
    else:
        return np.sum(np.abs(y-x)**p,axis=-1)
def minkowski_distance(x,y,p=2):
    """Compute the L**p distance between x and y"""
    x = np.asarray(x)
    y = np.asarray(y)
    if p==np.inf or p==1:
        return minkowski_distance_p(x,y,p)
    else:
        return minkowski_distance_p(x,y,p)**(1./p)

class Rectangle(object):
    """Hyperrectangle class.

    Represents a Cartesian product of intervals.
    """
    def __init__(self, maxes, mins):
        """Construct a hyperrectangle."""
        self.maxes = np.maximum(maxes,mins).astype(float)
        self.mins = np.minimum(maxes,mins).astype(float)
        self.m, = self.maxes.shape

    def __repr__(self):
        return "<Rectangle %s>" % zip(self.mins, self.maxes)

    def volume(self):
        """Total volume."""
        return np.prod(self.maxes-self.mins)

    def split(self, d, split):
        """Produce two hyperrectangles by splitting along axis d.

        In general, if you need to compute maximum and minimum
        distances to the children, it can be done more efficiently
        by updating the maximum and minimum distances to the parent.
        """ # FIXME: do this
        mid = np.copy(self.maxes)
        mid[d] = split
        less = Rectangle(self.mins, mid)
        mid = np.copy(self.mins)
        mid[d] = split
        greater = Rectangle(mid, self.maxes)
        return less, greater

    def min_distance_point(self, x, p=2.):
        """Compute the minimum distance between x and a point in the hyperrectangle."""
        return minkowski_distance(0, np.maximum(0,np.maximum(self.mins-x,x-self.maxes)),p)

    def max_distance_point(self, x, p=2.):
        """Compute the maximum distance between x and a point in the hyperrectangle."""
        return minkowski_distance(0, np.maximum(self.maxes-x,x-self.mins),p)

    def min_distance_rectangle(self, other, p=2.):
        """Compute the minimum distance between points in the two hyperrectangles."""
        return minkowski_distance(0, np.maximum(0,np.maximum(self.mins-other.maxes,other.mins-self.maxes)),p)

    def max_distance_rectangle(self, other, p=2.):
        """Compute the maximum distance between points in the two hyperrectangles."""
        return minkowski_distance(0, np.maximum(self.maxes-other.mins,other.maxes-self.mins),p)


class KDTree(object):
    """
    kd-tree for quick nearest-neighbor lookup

    This class provides an index into a set of k-dimensional points
    which can be used to rapidly look up the nearest neighbors of any
    point.

    The algorithm used is described in Maneewongvatana and Mount 1999.
    The general idea is that the kd-tree is a binary tree, each of whose
    nodes represents an axis-aligned hyperrectangle. Each node specifies
    an axis and splits the set of points based on whether their coordinate
    along that axis is greater than or less than a particular value.

    During construction, the axis and splitting point are chosen by the
    "sliding midpoint" rule, which ensures that the cells do not all
    become long and thin.

    The tree can be queried for the r closest neighbors of any given point
    (optionally returning only those within some maximum distance of the
    point). It can also be queried, with a substantial gain in efficiency,
    for the r approximate closest neighbors.

    For large dimensions (20 is already large) do not expect this to run
    significantly faster than brute force. High-dimensional nearest-neighbor
    queries are a substantial open problem in computer science.

    The tree also supports all-neighbors queries, both with arrays of points
    and with other kd-trees. These do use a reasonably efficient algorithm,
    but the kd-tree is not necessarily the best data structure for this
    sort of calculation.

    """

    def __init__(self, data, leafsize=10):
        """Construct a kd-tree.

        Parameters
        ----------
        data : array_like, shape (n,k)
            The data points to be indexed. This array is not copied, and
            so modifying this data will result in bogus results.
        leafsize : positive int
            The number of points at which the algorithm switches over to
            brute-force.
        """
        self.data = np.asarray(data)
        self.n, self.m = np.shape(self.data)
        self.leafsize = int(leafsize)
        if self.leafsize<1:
            raise ValueError("leafsize must be at least 1")
        self.maxes = np.amax(self.data,axis=0)
        self.mins = np.amin(self.data,axis=0)

        self.tree = self.__build(np.arange(self.n), self.maxes, self.mins)

    class node(object):
        if sys.version_info[0] >= 3:
            def __lt__(self, other): id(self) < id(other)
            def __gt__(self, other): id(self) > id(other)
            def __le__(self, other): id(self) <= id(other)
            def __ge__(self, other): id(self) >= id(other)
            def __eq__(self, other): id(self) == id(other)
    class leafnode(node):
        def __init__(self, idx):
            self.idx = idx
            self.children = len(idx)
    class innernode(node):
        def __init__(self, split_dim, split, less, greater):
            self.split_dim = split_dim
            self.split = split
            self.less = less
            self.greater = greater
            self.children = less.children+greater.children

    def __build(self, idx, maxes, mins):
        if len(idx)<=self.leafsize:
            return KDTree.leafnode(idx)
        else:
            data = self.data[idx]
            #maxes = np.amax(data,axis=0)
            #mins = np.amin(data,axis=0)
            d = np.argmax(maxes-mins)
            maxval = maxes[d]
            minval = mins[d]
            if maxval==minval:
                # all points are identical; warn user?
                return KDTree.leafnode(idx)
            data = data[:,d]

            # sliding midpoint rule; see Maneewongvatana and Mount 1999
            # for arguments that this is a good idea.
            split = (maxval+minval)/2
            less_idx = np.nonzero(data<=split)[0]
            greater_idx = np.nonzero(data>split)[0]
            if len(less_idx)==0:
                split = np.amin(data)
                less_idx = np.nonzero(data<=split)[0]
                greater_idx = np.nonzero(data>split)[0]
            if len(greater_idx)==0:
                split = np.amax(data)
                less_idx = np.nonzero(data<split)[0]
                greater_idx = np.nonzero(data>=split)[0]
            if len(less_idx)==0:
                # _still_ zero? all must have the same value
                assert np.all(data==data[0]), "Troublesome data array: %s" % data
                split = data[0]
                less_idx = np.arange(len(data)-1)
                greater_idx = np.array([len(data)-1])

            lessmaxes = np.copy(maxes)
            lessmaxes[d] = split
            greatermins = np.copy(mins)
            greatermins[d] = split
            return KDTree.innernode(d, split,
                    self.__build(idx[less_idx],lessmaxes,mins),
                    self.__build(idx[greater_idx],maxes,greatermins))

    def __query(self, x, k=1, eps=0, p=2, distance_upper_bound=np.inf):

        side_distances = np.maximum(0,np.maximum(x-self.maxes,self.mins-x))
        if p!=np.inf:
            side_distances**=p
            min_distance = np.sum(side_distances)
        else:
            min_distance = np.amax(side_distances)

        # priority queue for chasing nodes
        # entries are:
        #  minimum distance between the cell and the target
        #  distances between the nearest side of the cell and the target
        #  the head node of the cell
        q = [(min_distance,
              tuple(side_distances),
              self.tree)]
        # priority queue for the nearest neighbors
        # furthest known neighbor first
        # entries are (-distance**p, i)
        neighbors = []

        if eps==0:
            epsfac=1
        elif p==np.inf:
            epsfac = 1/(1+eps)
        else:
            epsfac = 1/(1+eps)**p

        if p!=np.inf and distance_upper_bound!=np.inf:
            distance_upper_bound = distance_upper_bound**p

        while q:
            min_distance, side_distances, node = heappop(q)
            if isinstance(node, KDTree.leafnode):
                # brute-force
                data = self.data[node.idx]
                ds = minkowski_distance_p(data,x[np.newaxis,:],p)
                for i in range(len(ds)):
                    if ds[i]<distance_upper_bound:
                        if len(neighbors)==k:
                            heappop(neighbors)
                        heappush(neighbors, (-ds[i], node.idx[i]))
                        if len(neighbors)==k:
                            distance_upper_bound = -neighbors[0][0]
            else:
                # we don't push cells that are too far onto the queue at all,
                # but since the distance_upper_bound decreases, we might get
                # here even if the cell's too far
                if min_distance>distance_upper_bound*epsfac:
                    # since this is the nearest cell, we're done, bail out
                    break
                # compute minimum distances to the children and push them on
                if x[node.split_dim]<node.split:
                    near, far = node.less, node.greater
                else:
                    near, far = node.greater, node.less

                # near child is at the same distance as the current node
                heappush(q,(min_distance, side_distances, near))

                # far child is further by an amount depending only
                # on the split value
                sd = list(side_distances)
                if p == np.inf:
                    min_distance = max(min_distance, abs(node.split-x[node.split_dim]))
                elif p == 1:
                    sd[node.split_dim] = np.abs(node.split-x[node.split_dim])
                    min_distance = min_distance - side_distances[node.split_dim] + sd[node.split_dim]
                else:
                    sd[node.split_dim] = np.abs(node.split-x[node.split_dim])**p
                    min_distance = min_distance - side_distances[node.split_dim] + sd[node.split_dim]

                # far child might be too far, if so, don't bother pushing it
                if min_distance<=distance_upper_bound*epsfac:
                    heappush(q,(min_distance, tuple(sd), far))

        if p==np.inf:
            return sorted([(-d,i) for (d,i) in neighbors])
        else:
            return sorted([((-d)**(1./p),i) for (d,i) in neighbors])

    def query(self, x, k=1, eps=0, p=2, distance_upper_bound=np.inf):
        """
        query the kd-tree for nearest neighbors

        Parameters
        ----------

        x : array-like, last dimension self.m
            An array of points to query.
        k : integer
            The number of nearest neighbors to return.
        eps : nonnegative float
            Return approximate nearest neighbors; the kth returned value
            is guaranteed to be no further than (1+eps) times the
            distance to the real kth nearest neighbor.
        p : float, 1<=p<=infinity
            Which Minkowski p-norm to use.
            1 is the sum-of-absolute-values "Manhattan" distance
            2 is the usual Euclidean distance
            infinity is the maximum-coordinate-difference distance
        distance_upper_bound : nonnegative float
            Return only neighbors within this distance. This is used to prune
            tree searches, so if you are doing a series of nearest-neighbor
            queries, it may help to supply the distance to the nearest neighbor
            of the most recent point.

        Returns
        -------

        d : array of floats
            The distances to the nearest neighbors.
            If x has shape tuple+(self.m,), then d has shape tuple if
            k is one, or tuple+(k,) if k is larger than one.  Missing
            neighbors are indicated with infinite distances.  If k is None,
            then d is an object array of shape tuple, containing lists
            of distances. In either case the hits are sorted by distance
            (nearest first).
        i : array of integers
            The locations of the neighbors in self.data. i is the same
            shape as d.

        Examples
        --------

        >>> from scipy.spatial import KDTree
        >>> x, y = np.mgrid[0:5, 2:8]
        >>> tree = KDTree(zip(x.ravel(), y.ravel()))
        >>> tree.data
        array([[0, 2],
               [0, 3],
               [0, 4],
               [0, 5],
               [0, 6],
               [0, 7],
               [1, 2],
               [1, 3],
               [1, 4],
               [1, 5],
               [1, 6],
               [1, 7],
               [2, 2],
               [2, 3],
               [2, 4],
               [2, 5],
               [2, 6],
               [2, 7],
               [3, 2],
               [3, 3],
               [3, 4],
               [3, 5],
               [3, 6],
               [3, 7],
               [4, 2],
               [4, 3],
               [4, 4],
               [4, 5],
               [4, 6],
               [4, 7]])
        >>> pts = np.array([[0, 0], [2.1, 2.9]])
        >>> tree.query(pts)
        (array([ 2.        ,  0.14142136]), array([ 0, 13]))

        """
        x = np.asarray(x)
        if np.shape(x)[-1] != self.m:
            raise ValueError("x must consist of vectors of length %d but has shape %s" % (self.m, np.shape(x)))
        if p<1:
            raise ValueError("Only p-norms with 1<=p<=infinity permitted")
        retshape = np.shape(x)[:-1]
        if retshape!=():
            if k is None:
                dd = np.empty(retshape,dtype=object)
                ii = np.empty(retshape,dtype=object)
            elif k>1:
                dd = np.empty(retshape+(k,),dtype=float)
                dd.fill(np.inf)
                ii = np.empty(retshape+(k,),dtype=int)
                ii.fill(self.n)
            elif k==1:
                dd = np.empty(retshape,dtype=float)
                dd.fill(np.inf)
                ii = np.empty(retshape,dtype=int)
                ii.fill(self.n)
            else:
                raise ValueError("Requested %s nearest neighbors; acceptable numbers are integers greater than or equal to one, or None")
            for c in np.ndindex(retshape):
                hits = self.__query(x[c], k=k, p=p, distance_upper_bound=distance_upper_bound)
                if k is None:
                    dd[c] = [d for (d,i) in hits]
                    ii[c] = [i for (d,i) in hits]
                elif k>1:
                    for j in range(len(hits)):
                        dd[c+(j,)], ii[c+(j,)] = hits[j]
                elif k==1:
                    if len(hits)>0:
                        dd[c], ii[c] = hits[0]
                    else:
                        dd[c] = np.inf
                        ii[c] = self.n
            return dd, ii
        else:
            hits = self.__query(x, k=k, p=p, distance_upper_bound=distance_upper_bound)
            if k is None:
                return [d for (d,i) in hits], [i for (d,i) in hits]
            elif k==1:
                if len(hits)>0:
                    return hits[0]
                else:
                    return np.inf, self.n
            elif k>1:
                dd = np.empty(k,dtype=float)
                dd.fill(np.inf)
                ii = np.empty(k,dtype=int)
                ii.fill(self.n)
                for j in range(len(hits)):
                    dd[j], ii[j] = hits[j]
                return dd, ii
            else:
                raise ValueError("Requested %s nearest neighbors; acceptable numbers are integers greater than or equal to one, or None")


    def __query_ball_point(self, x, r, p=2., eps=0):
        R = Rectangle(self.maxes, self.mins)

        def traverse_checking(node, rect):
            if rect.min_distance_point(x,p)>=r/(1.+eps):
                return []
            elif rect.max_distance_point(x,p)<r*(1.+eps):
                return traverse_no_checking(node)
            elif isinstance(node, KDTree.leafnode):
                d = self.data[node.idx]
                return node.idx[minkowski_distance(d,x,p)<=r].tolist()
            else:
                less, greater = rect.split(node.split_dim, node.split)
                return traverse_checking(node.less, less)+traverse_checking(node.greater, greater)
        def traverse_no_checking(node):
            if isinstance(node, KDTree.leafnode):

                return node.idx.tolist()
            else:
                return traverse_no_checking(node.less)+traverse_no_checking(node.greater)

        return traverse_checking(self.tree, R)

    def query_ball_point(self, x, r, p=2., eps=0):
        """Find all points within r of x

        Parameters
        ==========

        x : array_like, shape tuple + (self.m,)
            The point or points to search for neighbors of
        r : positive float
            The radius of points to return
        p : float 1<=p<=infinity
            Which Minkowski p-norm to use
        eps : nonnegative float
            Approximate search. Branches of the tree are not explored
            if their nearest points are further than r/(1+eps), and branches
            are added in bulk if their furthest points are nearer than r*(1+eps).

        Returns
        =======

        results : list or array of lists
            If x is a single point, returns a list of the indices of the neighbors
            of x. If x is an array of points, returns an object array of shape tuple
            containing lists of neighbors.


        Note: if you have many points whose neighbors you want to find, you may save
        substantial amounts of time by putting them in a KDTree and using query_ball_tree.
        """
        x = np.asarray(x)
        if x.shape[-1]!=self.m:
            raise ValueError("Searching for a %d-dimensional point in a %d-dimensional KDTree" % (x.shape[-1],self.m))
        if len(x.shape)==1:
            return self.__query_ball_point(x,r,p,eps)
        else:
            retshape = x.shape[:-1]
            result = np.empty(retshape,dtype=object)
            for c in np.ndindex(retshape):
                result[c] = self.__query_ball_point(x[c], r, p=p, eps=eps)
            return result

    def query_ball_tree(self, other, r, p=2., eps=0):
        """Find all pairs of points whose distance is at most r

        Parameters
        ==========

        other : KDTree
            The tree containing points to search against
        r : positive float
            The maximum distance
        p : float 1<=p<=infinity
            Which Minkowski norm to use
        eps : nonnegative float
            Approximate search. Branches of the tree are not explored
            if their nearest points are further than r/(1+eps), and branches
            are added in bulk if their furthest points are nearer than r*(1+eps).

        Returns
        =======

        results : list of lists
            For each element self.data[i] of this tree, results[i] is a list of the
            indices of its neighbors in other.data.
        """
        results = [[] for i in range(self.n)]
        def traverse_checking(node1, rect1, node2, rect2):
            if rect1.min_distance_rectangle(rect2, p)>r/(1.+eps):
                return
            elif rect1.max_distance_rectangle(rect2, p)<r*(1.+eps):
                traverse_no_checking(node1, node2)
            elif isinstance(node1, KDTree.leafnode):
                if isinstance(node2, KDTree.leafnode):
                    d = other.data[node2.idx]
                    for i in node1.idx:
                        results[i] += node2.idx[minkowski_distance(d,self.data[i],p)<=r].tolist()
                else:
                    less, greater = rect2.split(node2.split_dim, node2.split)
                    traverse_checking(node1,rect1,node2.less,less)
                    traverse_checking(node1,rect1,node2.greater,greater)
            elif isinstance(node2, KDTree.leafnode):
                less, greater = rect1.split(node1.split_dim, node1.split)
                traverse_checking(node1.less,less,node2,rect2)
                traverse_checking(node1.greater,greater,node2,rect2)
            else:
                less1, greater1 = rect1.split(node1.split_dim, node1.split)
                less2, greater2 = rect2.split(node2.split_dim, node2.split)
                traverse_checking(node1.less,less1,node2.less,less2)
                traverse_checking(node1.less,less1,node2.greater,greater2)
                traverse_checking(node1.greater,greater1,node2.less,less2)
                traverse_checking(node1.greater,greater1,node2.greater,greater2)

        def traverse_no_checking(node1, node2):
            if isinstance(node1, KDTree.leafnode):
                if isinstance(node2, KDTree.leafnode):
                    for i in node1.idx:
                        results[i] += node2.idx.tolist()
                else:
                    traverse_no_checking(node1, node2.less)
                    traverse_no_checking(node1, node2.greater)
            else:
                traverse_no_checking(node1.less, node2)
                traverse_no_checking(node1.greater, node2)

        traverse_checking(self.tree, Rectangle(self.maxes, self.mins),
                          other.tree, Rectangle(other.maxes, other.mins))
        return results

    def query_pairs(self, r, p=2., eps=0):
        """Find all pairs of points whose distance is at most r

        Parameters
        ==========

        r : positive float
            The maximum distance
        p : float 1<=p<=infinity
            Which Minkowski norm to use
        eps : nonnegative float
            Approximate search. Branches of the tree are not explored
            if their nearest points are further than r/(1+eps), and branches
            are added in bulk if their furthest points are nearer than r*(1+eps).

        Returns
        =======

        results : set
            set of pairs (i,j), i<j, for which the corresponing positions are
            close.

        """
        results = set()
        visited = set()
        def test_set_visited(node1, node2):
            i, j = sorted((id(node1),id(node2)))
            if (i,j) in visited:
                return True
            else:
                visited.add((i,j))
                return False
        def traverse_checking(node1, rect1, node2, rect2):
            if test_set_visited(node1, node2):
                return

            if id(node2)<id(node1):
                # This node pair will be visited in the other order
                #return
                pass

            if isinstance(node1, KDTree.leafnode):
                if isinstance(node2, KDTree.leafnode):
                    d = self.data[node2.idx]
                    for i in node1.idx:
                        for j in node2.idx[minkowski_distance(d,self.data[i],p)<=r]:
                            if i<j:
                                results.add((i,j))
                            elif j<i:
                                results.add((j,i))
                else:
                    less, greater = rect2.split(node2.split_dim, node2.split)
                    traverse_checking(node1,rect1,node2.less,less)
                    traverse_checking(node1,rect1,node2.greater,greater)
            elif isinstance(node2, KDTree.leafnode):
                less, greater = rect1.split(node1.split_dim, node1.split)
                traverse_checking(node1.less,less,node2,rect2)
                traverse_checking(node1.greater,greater,node2,rect2)
            elif rect1.min_distance_rectangle(rect2, p)>r/(1.+eps):
                return
            elif rect1.max_distance_rectangle(rect2, p)<r*(1.+eps):
                traverse_no_checking(node1.less, node2)
                traverse_no_checking(node1.greater, node2)
            else:
                less1, greater1 = rect1.split(node1.split_dim, node1.split)
                less2, greater2 = rect2.split(node2.split_dim, node2.split)
                traverse_checking(node1.less,less1,node2.less,less2)
                traverse_checking(node1.less,less1,node2.greater,greater2)
                traverse_checking(node1.greater,greater1,node2.less,less2)
                traverse_checking(node1.greater,greater1,node2.greater,greater2)

        def traverse_no_checking(node1, node2):
            if test_set_visited(node1, node2):
                return

            if id(node2)<id(node1):
                # This node pair will be visited in the other order
                #return
                pass
            if isinstance(node1, KDTree.leafnode):
                if isinstance(node2, KDTree.leafnode):
                    for i in node1.idx:
                        for j in node2.idx:
                            if i<j:
                                results.add((i,j))
                            elif j<i:
                                results.add((j,i))
                else:
                    traverse_no_checking(node1, node2.less)
                    traverse_no_checking(node1, node2.greater)
            else:
                traverse_no_checking(node1.less, node2)
                traverse_no_checking(node1.greater, node2)

        traverse_checking(self.tree, Rectangle(self.maxes, self.mins),
                          self.tree, Rectangle(self.maxes, self.mins))
        return results


    def count_neighbors(self, other, r, p=2.):
        """Count how many nearby pairs can be formed.

        Count the number of pairs (x1,x2) can be formed, with x1 drawn
        from self and x2 drawn from other, and where distance(x1,x2,p)<=r.
        This is the "two-point correlation" described in Gray and Moore 2000,
        "N-body problems in statistical learning", and the code here is based
        on their algorithm.

        Parameters
        ==========

        other : KDTree

        r : float or one-dimensional array of floats
            The radius to produce a count for. Multiple radii are searched with a single
            tree traversal.
        p : float, 1<=p<=infinity
            Which Minkowski p-norm to use

        Returns
        =======

        result : integer or one-dimensional array of integers
            The number of pairs. Note that this is internally stored in a numpy int,
            and so may overflow if very large (two billion).
        """

        def traverse(node1, rect1, node2, rect2, idx):
            min_r = rect1.min_distance_rectangle(rect2,p)
            max_r = rect1.max_distance_rectangle(rect2,p)
            c_greater = r[idx]>max_r
            result[idx[c_greater]] += node1.children*node2.children
            idx = idx[(min_r<=r[idx]) & (r[idx]<=max_r)]
            if len(idx)==0:
                return

            if isinstance(node1,KDTree.leafnode):
                if isinstance(node2,KDTree.leafnode):
                    ds = minkowski_distance(self.data[node1.idx][:,np.newaxis,:],
                                  other.data[node2.idx][np.newaxis,:,:],
                                  p).ravel()
                    ds.sort()
                    result[idx] += np.searchsorted(ds,r[idx],side='right')
                else:
                    less, greater = rect2.split(node2.split_dim, node2.split)
                    traverse(node1, rect1, node2.less, less, idx)
                    traverse(node1, rect1, node2.greater, greater, idx)
            else:
                if isinstance(node2,KDTree.leafnode):
                    less, greater = rect1.split(node1.split_dim, node1.split)
                    traverse(node1.less, less, node2, rect2, idx)
                    traverse(node1.greater, greater, node2, rect2, idx)
                else:
                    less1, greater1 = rect1.split(node1.split_dim, node1.split)
                    less2, greater2 = rect2.split(node2.split_dim, node2.split)
                    traverse(node1.less,less1,node2.less,less2,idx)
                    traverse(node1.less,less1,node2.greater,greater2,idx)
                    traverse(node1.greater,greater1,node2.less,less2,idx)
                    traverse(node1.greater,greater1,node2.greater,greater2,idx)
        R1 = Rectangle(self.maxes, self.mins)
        R2 = Rectangle(other.maxes, other.mins)
        if np.shape(r) == ():
            r = np.array([r])
            result = np.zeros(1,dtype=int)
            traverse(self.tree, R1, other.tree, R2, np.arange(1))
            return result[0]
        elif len(np.shape(r))==1:
            r = np.asarray(r)
            n, = r.shape
            result = np.zeros(n,dtype=int)
            traverse(self.tree, R1, other.tree, R2, np.arange(n))
            return result
        else:
            raise ValueError("r must be either a single value or a one-dimensional array of values")

    def sparse_distance_matrix(self, other, max_distance, p=2.):
        """Compute a sparse distance matrix

        Computes a distance matrix between two KDTrees, leaving as zero
        any distance greater than max_distance.

        Parameters
        ==========

        other : KDTree

        max_distance : positive float

        Returns
        =======

        result : dok_matrix
            Sparse matrix representing the results in "dictionary of keys" format.
        """
        result = scipy.sparse.dok_matrix((self.n,other.n))

        def traverse(node1, rect1, node2, rect2):
            if rect1.min_distance_rectangle(rect2, p)>max_distance:
                return
            elif isinstance(node1, KDTree.leafnode):
                if isinstance(node2, KDTree.leafnode):
                    for i in node1.idx:
                        for j in node2.idx:
                            d = minkowski_distance(self.data[i],other.data[j],p)
                            if d<=max_distance:
                                result[i,j] = d
                else:
                    less, greater = rect2.split(node2.split_dim, node2.split)
                    traverse(node1,rect1,node2.less,less)
                    traverse(node1,rect1,node2.greater,greater)
            elif isinstance(node2, KDTree.leafnode):
                less, greater = rect1.split(node1.split_dim, node1.split)
                traverse(node1.less,less,node2,rect2)
                traverse(node1.greater,greater,node2,rect2)
            else:
                less1, greater1 = rect1.split(node1.split_dim, node1.split)
                less2, greater2 = rect2.split(node2.split_dim, node2.split)
                traverse(node1.less,less1,node2.less,less2)
                traverse(node1.less,less1,node2.greater,greater2)
                traverse(node1.greater,greater1,node2.less,less2)
                traverse(node1.greater,greater1,node2.greater,greater2)
        traverse(self.tree, Rectangle(self.maxes, self.mins),
                 other.tree, Rectangle(other.maxes, other.mins))

        return result


def distance_matrix(x,y,p=2,threshold=1000000):
    """Compute the distance matrix.

    Computes the matrix of all pairwise distances.

    Parameters
    ==========

    x : array-like, m by k
    y : array-like, n by k
    p : float 1<=p<=infinity
        Which Minkowski p-norm to use.
    threshold : positive integer
        If m*n*k>threshold use a python loop instead of creating
        a very large temporary.

    Returns
    =======

    result : array-like, m by n


    """

    x = np.asarray(x)
    m, k = x.shape
    y = np.asarray(y)
    n, kk = y.shape

    if k != kk:
        raise ValueError("x contains %d-dimensional vectors but y contains %d-dimensional vectors" % (k, kk))

    if m*n*k <= threshold:
        return minkowski_distance(x[:,np.newaxis,:],y[np.newaxis,:,:],p)
    else:
        result = np.empty((m,n),dtype=float) #FIXME: figure out the best dtype
        if m<n:
            for i in range(m):
                result[i,:] = minkowski_distance(x[i],y,p)
        else:
            for j in range(n):
                result[:,j] = minkowski_distance(x,y[j],p)
        return result



# -*- coding: utf-8 -*-
#
# Defines meanshift grouper functions used by the group operation. Based on
# implementation of Francesco de Gasperin (see
# https://github.com/revoltek/LiLF/blob/master/lib_dd.py)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
import numpy as np
from . import _grouper
import logging
log = logging.getLogger('LSMTool.Group')


class Grouper(object):
    """
    Class to group a list of coordinates and fluxes by the meanshift algorighm.
    Based on: http://www.chioka.in/meanshift-algorithm-for-the-rest-of-us-python/
    Parameters
    ----------
    coords : list
        List of (x, y) pixel coordinates for source positions
    fluxes: list
        total flux density for each source
    kernel_size : float
        Kernel size in pixels
    n_iterations : int
        Number of iterations
    look_distance : float
        Look distance in pixels
    grouping_distance : float
        Grouping distance in pixels
    """
    def __init__(self, coords, fluxes, kernel_size, n_iterations, look_distance,
                 grouping_distance):
        self.coords = np.array(coords)
        self.fluxes = fluxes
        self.kernel_size = kernel_size
        self.n_iterations = n_iterations
        self.look_distance = look_distance
        self.grouping_distance = grouping_distance
        self.past_coords = [np.copy(self.coords)]
        self.clusters = []
        self.g = _grouper.Grouper()
        self.g.readCoordinates(self.coords, self.fluxes)
        self.g.setKernelSize(self.kernel_size)
        self.g.setNumberOfIterations(self.n_iterations)
        self.g.setLookDistance(self.look_distance)
        self.g.setGroupingDistance(self.grouping_distance)


    def run(self):
        self.g.run()

    def grouping(self):
        """
        Take the last coords set and group sources nearby, then return a list of lists.
        Each list has the index of one cluster.
        """

        self.g.group(self.clusters)

        log.info('Creating %i groups.' % len(self.clusters))
        return self.clusters



from __future__ import print_function
from setuptools import setup, Command, Extension, Distribution
from setuptools.command.build_ext import build_ext
import os
import sys


# Functions read() and get_version() were copied from Pip package.
# Purpose is to get version info from current package without it
# being installed (which is usually the case when setup.py is run).
def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    # intentionally *not* adding an encoding option to open, See:
    #   https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
    with open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            # __version__ = "0.9"
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")


# Flag that determines whether to build the optional (but faster) C++
# extensions. Set to False to install only the pure Python versions.
if "--build_c_extentions" in sys.argv:
    build_c_extentions = True
    sys.argv.remove("--build_c_extentions")
else:
    build_c_extentions = False

# Handle Python 3-only dependencies
if sys.version_info < (3, 0):
    reqlist = ['numpy', 'astropy >= 0.4, <3.0']
else:
    # Require astropy v3.2 or later to get much faster copies
    reqlist = ['numpy', 'astropy >= 3.2', 'matplotlib', 'scipy', 'everybeam']
if build_c_extentions:
    reqlist.append('pybind11>=2.2.0')
    ext_modules = [Extension('lsmtool.operations._grouper',
                             ['lsmtool/operations/_grouper.cpp'],
                             language='c++')]
else:
    ext_modules = []


class PyTest(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        import subprocess
        errno = subprocess.call([sys.executable, 'runtests.py'])
        raise SystemExit(errno)


class LSMToolDistribution(Distribution):

    def is_pure(self):
        return self.pure

    def has_ext_modules(self):
        return not self.pure

    global_options = Distribution.global_options + [
        ('pure', None, "use pure Python code instead of C++ extensions")]
    pure = False


class BuildExt(build_ext):

    def build_extensions(self):
        opts = ['-std=c++11']
        if sys.platform == 'darwin':
            opts += ['-stdlib=libc++']
        else:
            # Enable OpenMP support only on non-Darwin platforms, as clang
            # does not support it without some work
            opts += ['-fopenmp']
        for ext in self.extensions:
            ext.extra_compile_args = opts
        build_ext.build_extensions(self)

setup(
    name='lsmtool',
    version=get_version("lsmtool/_version.py"),
    url='https://git.astron.nl/RD/LSMTool',
    project_urls={
        "Documentation": "https://lsmtool.readthedocs.io/en/latest/",
        "Source": "https://git.astron.nl/RD/LSMTool"
    },
    author='David Rafferty',
    author_email='drafferty@hs.uni-hamburg.de',
    description='The LOFAR Local Sky Model Tool',
    long_description=read("README.md"),
    long_description_content_type='text/markdown',
    license="GPL",
    platforms='any',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python',
        'Natural Language :: English',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Astronomy',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    install_requires=reqlist,
    scripts=['bin/lsmtool', 'bin/correct_gaussian_orientation.py'],
    ext_modules=ext_modules,
    cmdclass={'build_ext': BuildExt},
    distclass=LSMToolDistribution,
    packages=['lsmtool', 'lsmtool.operations'],
    setup_requires=['pytest-runner'],
    tests_require=['pytest']
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# This operation implements transferring of patches from one sky model to another
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import logging

log = logging.getLogger('LSMTool.TRANSFER')
log.debug('Loading TRANSFER module.')


def run(step, parset, LSM):

    patchFile = parset.getString('.'.join(["LSMTool.Steps", step, "PatchFile"]), '' )
    outFile = parset.getString('.'.join(["LSMTool.Steps", step, "OutFile"]), '' )

    try:
        transfer(LSM, patchFile)
        result = 0
    except Exception as e:
        log.error(e)
        result = 1

    # Write to outFile
    if outFile != '' and result == 0:
        LSM.write(outFile, clobber=True)

    return result


def transfer(LSM, patchSkyModel, matchBy='name', radius=0.1):
    """
    Transfer patches from the input sky model.

    Sources matching those in patchSkyModel will be grouped into
    the patches defined in patchSkyModel. Sources that do not appear in
    patchSkyModel will be placed into separate patches (one per source).
    Patch positions are not transferred (as they may no longer be appropriate
    after transfer).

    Parameters
    ----------
    LSM : SkyModel object
        Input sky model
    patchSkyModel : str or SkyModel object
        Input sky model from which to transfer patches
    matchBy : str, optional
        Determines how matching sources are determined:
        - 'name' => matches are identified by name
        - 'position' => matches are identified by radius. Sources within the
            radius specified by the radius parameter are considered matches
    radius : float or str, optional
        Radius in degrees (if float) or 'value unit' (if str; e.g., '30 arcsec')
        for matching when matchBy='position'

    Examples
    --------
    Transfer patches from one sky model to another by matching to the source
    names and set their positions::

        >>> LSM = lsmtool.load('sky.model')
        >>> transfer(LSM, 'master_sky.model')
        >>> setPatchPositions(LSM, method='mid')

    Transfer patches by matching sources that lie within 10 arcsec of one
    another::

        >>> LSM = lsmtool.load('sky.model')
        >>> transfer(LSM, 'master_sky.model', matchBy='position', radius='10.0 arcsec')
        >>> setPatchPositions(LSM, method='mid')

    """
    from ..skymodel import SkyModel
    from ..operations_lib import matchSky

    if len(LSM) == 0:
        log.error('Sky model is empty.')
        return

    if type(patchSkyModel) is str:
        masterLSM = SkyModel(patchSkyModel)
    else:
        masterLSM = patchSkyModel

    # Group LSM by source. This ensures that any sources not in the master
    # sky model are given a patch of their own
    log.debug('Grouping master sky model by one patch per source...')
    LSM.group('every')
    patchNames = LSM.getColValues('Patch')
    masterPatchNames = masterLSM.getColValues('Patch')
    table = LSM.table.copy()

    if matchBy.lower() == 'name':
        log.debug('Transferring patches by matching names...')
        names = LSM.getColValues('Name')
        masterNames = masterLSM.getColValues('Name')

        lrIntr = lambda l, r: list(set(l).intersection(r))
        commonNames = lrIntr(names.tolist(), masterNames.tolist())
        nMissing = len(names) - len(commonNames)

        for name in commonNames:
            indx = LSM.getRowIndex(name)
            masterIndx = masterLSM.getRowIndex(name)
            table['Patch'][indx] = masterLSM.table['Patch'][masterIndx]

    elif matchBy.lower() == 'position':
        log.debug('Transferring patches by matching positions...')
        matches1, matches2 = matchSky(LSM, masterLSM, radius=radius)
        nMissing = len(LSM) - len(matches1)

        # Set patch names to be the same for the matches
        table['Patch'][matches1] = masterLSM.table['Patch'][matches2]

    log.debug('Number of sources not present in patchSkyModel: {0}'.format(
        nMissing))
    LSM.table = table
    LSM._updateGroups()
    LSM._addHistory('TRANSFER')
    LSM._info()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# This operation implements plotting of sources
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import logging

log = logging.getLogger('LSMTool.PLOT')
log.debug('Loading PLOT module.')


def run(step, parset, LSM):

    outFile = parset.getString('.'.join(["LSMTool.Steps", step, "OutFile"]), '')
    labelBy = parset.getString('.'.join(["LSMTool.Steps", step, "LabelBy"]), '')

    if outFile == '':
        outFile = None
    if labelBy == '':
        labelBy = None

    try:
        plot(LSM, fileName=outFile, labelBy=labelBy)
        result = 0
    except Exception as e:
        log.error(e)
        result = 1

    return 0


def plot(LSM, fileName=None, labelBy=None):
    """
    Shows a simple plot of the sky model.

    The circles in the plot are scaled with flux. If the sky model is grouped
    into patches, sources are colored by patch and the patch positions are
    indicated with stars.

    Parameters
    ----------
    LSM : SkyModel object
        Input sky model
    fileName : str, optional
        If given, the plot is saved to a file instead of displayed
    labelBy : str, optional
        One of 'source' or 'patch': label points using source names ('source') or
        patch names ('patch')

    Examples:
    ---------
    Plot and display to the screen::

        >>> LSM = lsmtool.load('sky.model')
        >>> plot(LSM)

    Plot and save to a PDF file::

        >>> plot(LSM, 'sky_plot.pdf')

    """
    try:
        import os
        if 'DISPLAY' not in os.environ:
            import matplotlib
            if matplotlib.get_backend() != 'Agg':
                matplotlib.use("Agg")
        import matplotlib.pyplot as plt
        from matplotlib.ticker import FuncFormatter
    except Exception as e:
        raise ImportError('PyPlot could not be imported. Plotting is not '
            'available: {0}'.format(e))
    try:
        try:
            from astropy.visualization.wcsaxes import WCSAxes
            hasWCSaxes = True
        except:
            from wcsaxes import WCSAxes
            hasWCSaxes = True
    except:
        hasWCSaxes = False
    import numpy as np
    from ..operations_lib import radec2xy, makeWCS
    global midRA, midDec, ymin, xmin

    if len(LSM) == 0:
        log.error('Sky model is empty.')
        return

    fig = plt.figure(1,figsize=(7.66,7))
    plt.clf()
    x, y, midRA, midDec  = LSM._getXY()
    if hasWCSaxes:
        wcs = makeWCS(midRA, midDec)
        ax = WCSAxes(fig, [0.16, 0.1, 0.8, 0.8], wcs=wcs)
        fig.add_axes(ax)
    else:
        ax = plt.gca()
    if LSM.hasPatches:
        nsrc = len(LSM.getPatchNames())
    else:
        nsrc = len(LSM)
    sm = plt.cm.ScalarMappable(cmap=plt.cm.Set3, norm=plt.Normalize(vmin=0,
        vmax=nsrc))
    sm._A = []

    # Set symbol sizes by flux, making sure no symbol is smaller than 50 or
    # larger than 1000
    s = []
    fluxes = LSM.getColValues('I')
    if len(fluxes[fluxes > 0.0]) == 0:
        minflux = 0.0
    else:
        minflux = np.min(fluxes[fluxes > 0.0])
    for flux in LSM.getColValues('I'):
        if flux > 0.0:
            s.append(min(1000.0, (1.0+2.0*np.log10(flux/minflux))*50.0))
        else:
            s.append(50.0)

    # Color sources by patch if grouped
    c = [0]*len(LSM)
    cp = []
    if LSM.hasPatches:
        for p, patchName in enumerate(LSM.getPatchNames()):
            indices = LSM.getRowIndex(patchName)
            cp.append(sm.to_rgba(p))
            for ind in indices:
                c[ind] = sm.to_rgba(p)
    else:
        c = [sm.to_rgba(0)] * nsrc

    # Plot sources
    if hasWCSaxes:
        RA = LSM.getColValues('Ra')
        Dec = LSM.getColValues('Dec')
        ax.set_xlim(np.min(x)-20, np.max(x)+20)
        ax.set_ylim(np.min(y)-20, np.max(y)+20)
    plt.scatter(x, y, s=s, c=c)

    if LSM.hasPatches:
        RAp, Decp = LSM.getPatchPositions(asArray=True)
        goodInd = np.where( (RAp != 0.0) & (Decp != 0.0) )
        if len(goodInd[0]) < len(RAp):
            log.info('Some patch positions are unset. Run setPatchPositions() '
                'before plotting to see patch positions and patch names.')
        xp, yp = radec2xy(RAp[goodInd], Decp[goodInd], midRA, midDec)
        plt.scatter(xp, yp, s=100, c=cp, marker='*')

    # Set axis labels, etc.
    if hasWCSaxes:
        RAAxis = ax.coords['ra']
        RAAxis.set_axislabel('RA', minpad=0.75)
        RAAxis.set_major_formatter('hh:mm:ss')
        DecAxis = ax.coords['dec']
        DecAxis.set_axislabel('Dec', minpad=0.75)
        DecAxis.set_major_formatter('dd:mm:ss')
        ax.coords.grid(color='black', alpha=0.5, linestyle='solid')
    else:
        plt.xlabel("RA (arb. units)")
        plt.ylabel("Dec (arb. units)")

    if labelBy is not None:
        if labelBy.lower() == 'source':
            labels = LSM.getColValues('name')
            xls = x
            yls = y
        elif labelBy.lower() == 'patch':
            if LSM.hasPatches:
                labels = LSM.getPatchNames()
                xls = xp
                yls = yp
            else:
                labels = LSM.getColValues('name')
                xls = x
                yls = y
        else:
            raise ValueError("The lableBy parameter must be one of 'source' or "
                "'patch'.")
        for label, xl, yl in zip(labels, xls, yls):
            plt.annotate(label, xy = (xl, yl), xytext = (-2, 2), textcoords=
                'offset points', ha='right', va='bottom')

    # Define coodinate formater to show RA and Dec under mouse pointer
    RAformatter = FuncFormatter(RAtickformatter)
    ax.format_coord = formatCoord

    if fileName is not None:
        plt.savefig(fileName)
    else:
        plt.show()
    plt.close(fig)


def formatCoord(x, y):
    """Custom coordinate format"""
    from ..operations_lib import xy2radec

    global midRA, midDec
    RA, Dec = xy2radec([x], [y], midRA, midDec)
    return 'RA = {0:.2f} Dec = {1:.2f}'.format(RA[0], Dec[0])


def RAtickformatter(x, pos):
    """Changes x tick labels from pixels to RA in degrees"""
    from ..operations_lib import xy2radec

    global ymin, midRA, midDec
    ratick = xy2radec([x], [ymin], midRA, midDec)[0][0]
    rastr = '{0:.2f}'.format(ratick)
    return rastr


def Dectickformatter(y, pos):
    """Changes y tick labels from pixels to Dec in degrees"""
    from ..operations_lib import xy2radec

    global xmin, midRA, midDec
    dectick = xy2radec([xmin], [y], midRA, midDec)[1][0]
    decstr = '{0:.2f}'.format(dectick)
    return decstr




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# This operation implements grouping of sources into patches
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import logging

log = logging.getLogger('LSMTool.GROUP')
log.debug('Loading GROUP module.')


def run(step, parset, LSM):

    outFile = parset.getString('.'.join(["LSMTool.Steps", step, "OutFile"]), '' )
    algorithm = parset.getString('.'.join(["LSMTool.Steps", step, "Algorithm"]), 'single' )
    root = parset.getString('.'.join(["LSMTool.Steps", step, "Root"]), 'Patch' )
    targetFlux = parset.getString('.'.join(["LSMTool.Steps", step, "TargetFlux"]), '1.0 Jy' )
    numClusters = parset.getInt('.'.join(["LSMTool.Steps", step, "NumClusters"]), 100 )
    threshold = parset.getString('.'.join(["LSMTool.Steps", step, "Threshold"]), '1.0 Jy' )
    FWHM = parset.getString('.'.join(["LSMTool.Steps", step, "FWHM"]), None )
    applyBeam = parset.getBool('.'.join(["LSMTool.Steps", step, "ApplyBeam"]), False )
    method = parset.getString('.'.join(["LSMTool.Steps", step, "Method"]), 'mid' )
    pad_index = parset.getBool('.'.join(["LSMTool.Steps", step, "PadIndex"]), False )
    byPatch = parset.getBool('.'.join(["LSMTool.Steps", step, "ByPatch"]), False )
    facet = parset.getString('.'.join(["LSMTool.Steps", step, "Facet"]), '' )
    kernelSize = parset.getString('.'.join(["LSMTool.Steps", step, "KernelSize"]), '0.1' )
    nIterations = parset.getInt('.'.join(["LSMTool.Steps", step, "NIterations"]), '100' )
    lookDistance = parset.getString('.'.join(["LSMTool.Steps", step, "LookDistance"]), '0.2' )
    groupingDistance = parset.getString('.'.join(["LSMTool.Steps", step, "GroupingDistance"]), '0.01' )

    try:
        group(LSM, algorithm, targetFlux, numClusters, FWHM, threshold, applyBeam, root,
              pad_index, method, facet, byPatch, float(kernelSize), nIterations,
              float(lookDistance), float(groupingDistance))
        result = 0
    except Exception as e:
        log.error(e)
        result = 1

    # Write to outFile
    if outFile != '' and result == 0:
        LSM.write(outFile, clobber=True)

    return result


def group(LSM, algorithm, targetFlux=None, weightBySize=False, numClusters=100, FWHM=None,
          threshold=0.1, applyBeam=False, root='Patch', pad_index=False, method='mid',
          facet="", byPatch=False, kernelSize=0.1, nIterations=100, lookDistance=0.2,
          groupingDistance=0.01):
    """
    Groups sources into patches.

    Parameters
    ----------
    LSM : SkyModel object
        Input sky model
    algorithm : str
        Algorithm to use for grouping:
        - 'single' => all sources are grouped into a single patch
        - 'every' => every source gets a separate patch named 'source_patch'
        - 'cluster' => SAGECAL clustering algorithm that groups sources into
            specified number of clusters (specified by the numClusters parameter)
        - 'tessellate' => group into tiles whose total flux approximates
            the target flux (specified by the targetFlux parameter)
        - 'threshold' => group by convolving the sky model with a Gaussian beam
            and then thresholding to find islands of emission (NOTE: all sources
            are currently considered to be point sources of flux unity)
        - 'facet' => group by facets using as an input a fits file. It requires
            the use of the additional parameter 'facet' to enter the name of the
            fits file.
        - 'voronoi' => given a previously grouped sky model, voronoi tesselate
            using the patch positions for patches above the target flux (specified
            by the targetFlux parameter)
        - 'meanshift' => use the meanshift clustering algorithm
        - the filename of a mask image => group by masked regions (where mask =
            True). Sources outside of masked regions are given patches of their
            own
    targetFlux : str or float, optional
        Target flux for tessellation (the total flux of each tile will be close
        to this value) and voronoi algorithms. The target flux can be specified
        as either a float in Jy or as a string with units (e.g., '25.0 mJy')
    weightBySize : bool, optional
        If True, fluxes are weighted by patch size (as median_size / size) when
        the targetFlux criterion is applied. Patches with sizes below the median
        (flux-weighted) size are upweighted and those above the median are
        downweighted
    numClusters : int, optional
        Number of clusters for clustering. Sources are grouped around the
        numClusters brightest sources
    FWHM : str or float, optional
        FWHM of convolving Gaussian used for thresholding. The FWHM can
        be specified as either a float in degrees or as a string with units
        (e.g., '25.0 arcsec')
    threshold : float, optional
        Value between 0 and 1 above which emission is considered for thresholding
    applyBeam : bool, optional
        If True, fluxes will be attenuated by the beam
    root : str, optional
        Root string from which patch names are constructed. For 'single', the
        patch name will be set to root; for the other grouping algorithms, the
        patch names will be 'root_INDX', where INDX is an integer ranging from
        (0:nPatches)
    pad_index : bool, optional
        If True, pad the INDX used in the patch names. E.g., facet_patch_001
        instead of facet_patch_1
    method : None or str, optional
        This parameter specifies the method used to set the patch positions:
        - 'mid' => the position is set to the midpoint of the patch
        - 'mean' => the positions is set to the mean RA and Dec of the patch
        - 'wmean' => the position is set to the flux-weighted mean RA and
        Dec of the patch
        - 'zero' => set all positions to [0.0, 0.0]
    facet : str, optional
        Facet fits file used with the algorithm 'facet'
    byPatch : bool, optional
        For the 'tessellate' or 'meanshift' algorithms, use patches instead of sources
    kernelSize : float, optional
        Kernel size in degrees for meanshift grouping
    nIterations : int, optional
        Number of iterations for meanshift grouping
    lookDistance : float, optional
        Look distance in degrees for meanshift grouping
    groupingDistance : float, optional
        Grouping distance in degrees for meanshift grouping

    Examples
    --------
    Tesselate the sky model into patches with approximately 30 Jy total
    flux:

        >>> LSM = lsmtool.load('sky.model')
        >>> group(LSM, 'tessellate', targetFlux=30.0)

    """
    from . import _tessellate
    from . import _cluster
    from . import _threshold
    try:
        from . import _meanshiftc as _meanshift
    except ImportError:
        from . import _meanshift
    import numpy as np
    import os
    from itertools import groupby

    if len(LSM) == 0:
        log.error('Sky model is empty.')
        return

    if algorithm.lower() == 'single':
        LSM.ungroup()
        addSingle(LSM, root)

    elif algorithm.lower() == 'every':
        LSM.ungroup()
        addEvery(LSM)

    elif algorithm.lower() == 'cluster':
        LSM.ungroup()
        patches = _cluster.compute_patch_center(LSM, applyBeam=applyBeam)
        patchCol = _cluster.create_clusters(LSM, patches, numClusters,
                                            applyBeam=applyBeam, root=root,
                                            pad_index=pad_index)
        LSM.setColValues('Patch', patchCol, index=2)

    elif algorithm.lower() == 'tessellate':
        if targetFlux is None:
            raise ValueError('Please specify the targetFlux parameter.')
        else:
            units = 'Jy'
            if type(targetFlux) is str:
                parts = [''.join(g).strip() for _, g in groupby(targetFlux, str.isalpha)]
                targetFlux = float(parts[0])
                if len(parts) == 2:
                    units = parts[1]
        if byPatch:
            if 'Patch' not in LSM.table.keys():
                raise ValueError('Sky model must be grouped before "byPatch" can be used.')
            x, y, midRA, midDec = LSM._getXY(byPatch=True)
            f = LSM.getColValues('I', units=units, applyBeam=applyBeam, aggregate='sum')
        else:
            LSM.ungroup()
            x, y, midRA, midDec = LSM._getXY()
            f = LSM.getColValues('I', units=units, applyBeam=applyBeam)
        vobin = _tessellate.bin2D(np.array(x), np.array(y), f, target_flux=targetFlux)
        try:
            vobin.bin_voronoi()
            patchCol = _tessellate.bins2Patches(vobin, root=root, pad_index=pad_index)
            if byPatch:
                newPatchNames = patchCol.copy()
                origPatchNames = LSM.getPatchNames()
                patchCol = np.zeros(len(LSM), dtype='U100')
                for newPatchName, origPatchName in zip(newPatchNames, origPatchNames):
                    ind = np.array(LSM.getRowIndex(origPatchName))
                    patchCol[ind] = newPatchName
                LSM.setColValues('Patch', patchCol, index=2)
            else:
                LSM.setColValues('Patch', patchCol, index=2)
        except ValueError:
            # Catch error in some cases with high target flux relative to
            # total model flux
            addSingle(LSM, root+'_0')

    elif algorithm.lower() == 'threshold':
        from astropy.coordinates import SkyCoord, Angle

        if threshold is None:
            threshold = 0.1
        if threshold < 0.01:
            threshold = 0.01
        if threshold > 1.0:
            threshold = 1.0
        if FWHM is None:
            raise ValueError('Please specify the FWHM parameter.')
        else:
            units = 'degree'
            if type(FWHM) is str:
                parts = [''.join(g).strip() for _, g in groupby(FWHM, str.isalpha)]
                FWHM = float(parts[0])
                if len(parts) == 2:
                    units = parts[1]
            fwhmArcsec = Angle(FWHM, unit=units).to('arcsec').value

        patchCol = _threshold.getPatchNamesByThreshold(LSM, fwhmArcsec, threshold,
            root=root, pad_index=pad_index)
        LSM.setColValues('Patch', patchCol, index=2)

    elif algorithm.lower() == 'voronoi':
        from astropy.coordinates import SkyCoord
        import astropy.units as u

        if 'Patch' not in LSM.table.keys():
            raise ValueError('Sky model must be grouped before "voronoi" can be used.')
        else:
            dirs = LSM.getPatchPositions()

        if targetFlux is not None:
            # Select only those patches that lie above the target flux
            units = 'Jy'
            if type(targetFlux) is str:
                parts = [''.join(g).strip() for _, g in groupby(targetFlux, str.isalpha)]
                targetFlux = float(parts[0])
                if len(parts) == 2:
                    units = parts[1]
            dirs_names = []
            names = LSM.getPatchNames()
            fluxes = LSM.getColValues('I', aggregate='sum', units=units, applyBeam=applyBeam)
            if weightBySize:
                sizes = LSM.getPatchSizes(units='arcsec', weight=True, applyBeam=applyBeam)
                sizes[sizes < 1.0] = 1.0
                bright_ind = np.where(fluxes >= targetFlux)
                medianSize = np.median(sizes[bright_ind])
                weights = medianSize / sizes
                weights[weights > 1.0] = 1.0
                weights[weights < 0.5] = 0.5
                fluxes *= weights
            for name, flux in zip(names, fluxes):
                if flux >= targetFlux:
                    dirs_names.append(name)
            if len(dirs_names) == 0:
                log.warn('No patches meet specified targetFlux. All sources placed in a single patch.')
                LSM.ungroup()
                addSingle(LSM, root)
                return 0
        else:
            # Use all patches
            dirs_names = [name for name, d in dirs.items()]

        dirs_ras = []
        dirs_decs = []
        for name in dirs_names:
            d = dirs[name]
            dirs_ras.append(d[0])
            dirs_decs.append(d[1])
        RADeg = LSM.getColValues('Ra', units='degree')
        DecDeg = LSM.getColValues('Dec', units='degree')
        patchNames = []
        for r, d in zip(RADeg, DecDeg):
            dists = SkyCoord(r*u.degree, d*u.degree).separation(
                SkyCoord(dirs_ras*u.degree, dirs_decs*u.degree))
            patchNames.append(dirs_names[np.argmin(dists)])
        LSM.setColValues('Patch', patchNames, index=2)

    elif algorithm.lower() == 'meanshift':
        if byPatch:
            if 'Patch' not in LSM.table.keys():
                raise ValueError('Sky model must be grouped before "byPatch" can be used.')
            x, y, midRA, midDec = LSM._getXY(byPatch=True)
            f = LSM.getColValues('I', applyBeam=applyBeam, aggregate='sum')
        else:
            addEvery(LSM)
            x, y, midRA, midDec = LSM._getXY()
            f = LSM.getColValues('I', applyBeam=applyBeam)
        crdelt = 0.066667  # WCS delta in deg/pixel, as used by LSM._getXY()
        grouper = _meanshift.Grouper(list(zip(x, y)), f, kernelSize/crdelt, nIterations,
                                     lookDistance/crdelt, groupingDistance/crdelt)
        grouper.run()
        clusters = grouper.grouping()
        patchNames = LSM.getPatchNames()
        table = LSM.table.copy()
        for cindx, cluster in enumerate(clusters):
            if pad_index:
                name = '{0}_{1}'.format(root, str(int(cindx)).zfill(int(np.ceil(np.log10(len(cluster)+1)))))
            else:
                name = '{0}_{1}'.format(root, str(int(cindx)))
            patches = patchNames[cluster]
            for patchName in patches:
                indices = LSM.getRowIndex(patchName)
                table['Patch'][indices] = name
        LSM.table = table
        LSM._updateGroups()

    elif algorithm.lower() == 'facet':
        if os.path.exists(facet):
            RADeg = LSM.getColValues('Ra', units='degree')
            DecDeg = LSM.getColValues('Dec', units='degree')
            facet_col = get_facet_values(facet, RADeg, DecDeg, root=root, pad_index=pad_index)
            LSM.setColValues('Patch', facet_col, index=2)
        else:
            raise ValueError('Please enter the facet filename in the facet parameter.')

    elif os.path.exists(algorithm):
        # Mask image
        mask = algorithm
        RARad = LSM.getColValues('Ra', units='radian')
        DecRad = LSM.getColValues('Dec', units='radian')
        patchCol = getPatchNamesFromMask(mask, RARad, DecRad, root=root, pad_index=pad_index)
        LSM.setColValues('Patch', patchCol, index=2)

    else:
        raise ValueError('Grouping alogrithm not understood.')

    # Update table grouping and set default patch positions
    LSM._updateGroups()
    history = "algorithm = '{0}'".format(algorithm)
    if algorithm.lower() == 'cluster':
        history += ', numClusters = {0}'.format(numClusters)
    elif algorithm.lower() == 'tessellate' or algorithm.lower() == 'voronoi':
        history += ', targetFlux = {0}'.format(targetFlux)
    LSM._addHistory("GROUP ({0})".format(history))
    LSM.setPatchPositions(method=method, applyBeam=applyBeam)
    LSM._info()
    return 0


def addSingle(LSM, patchName):
    """Add a Patch column with a single patch name"""
    import numpy as np

    uniformCol = np.array([patchName]*len(LSM.table))
    LSM.setColValues('Patch', uniformCol, index=2)


def addEvery(LSM):
    """Add a Patch column with a different name for each source"""
    names = LSM.getColValues('Name').copy()
    for i, name in enumerate(names):
        names[i] = name + '_patch'
    LSM.setColValues('Patch', names, index=2)


def getPatchNamesFromMask(mask, RARad, DecRad, root='mask', pad_index=False):
    """
    Returns an array of patch names for each (RA, Dec) pair in radians
    """
    import pyrap.images as pim
    import scipy.ndimage as nd
    import numpy as np

    maskdata = pim.image(mask)
    maskval = maskdata.getdata()[0][0]

    act_pixels = maskval
    rank = len(act_pixels.shape)
    connectivity = nd.generate_binary_structure(rank, rank)
    mask_labels, count = nd.label(act_pixels, connectivity)

    patchNums = []
    patchNames = []
    for raRad, decRad in zip(RARad, DecRad):
        (a, b, _, _) = maskdata.toworld([0, 0, 0, 0])
        (_, _, pixY, pixX) = maskdata.topixel([a, b, decRad, raRad])
        try:
            # != is a XOR for booleans
            patchNums.append(mask_labels[int(pixY), int(pixX)])
        except:
            patchNums.append(0)

    # Check if there is a patch with id = 0. If so, this means there were
    # some Gaussians that fell outside of the regions in the patch
    # mask file.
    n = 0
    for p in patchNums:
        if p != 0:
            if pad_index:
                patchNames.append('{0}_patch_'.format(root) +
                                  str(p).zfill(int(np.ceil(np.log10(len(set(patchNums))+1)))))
            else:
                patchNames.append('{0}_patch_'.format(root)+str(p))
        else:
            patchNames.append('patch_'+str(n))
            n += 1

    return np.array(patchNames)


def get_facet_values(facet, ra, dec, root="facet", default=0, pad_index=False):
    """
    Extract the value from a fits facet file
    """
    import numpy as np
    from astropy.io import fits
    from astropy.wcs import WCS

    # TODO: Check astropy version
    # TODO: Check facet is a fits file

    with fits.open(facet) as f:
        shape = f[0].data.shape

        w = WCS(f[0].header)
        if len(w.wcs.crval) == 4:
            freq = w.wcs.crval[2]
            stokes = w.wcs.crval[3]
            xe, ye, _1, _2 = w.all_world2pix(ra, dec, freq, stokes, 1)
        elif len(w.wcs.crval) == 2:
            xe, ye = w.all_world2pix(ra, dec, 1)
        else:
            raise ValueError('Input mask must have 2 axes (x, y) or 4 axes (x, y, freq, stokes).')
        x, y = np.round(xe).astype(int), np.round(ye).astype(int)

        # Dummy value for points out of the fits area
        x[(x < 0) | (x >= shape[-1])] = -1
        y[(y < 0) | (y >= shape[-2])] = -1

        if len(w.wcs.crval) == 4:
            data = f[0].data[0, 0, :, :]
        else:
            data = f[0].data[:, :]

        values = data[y, x]

        # Assign the default value to NaNs and points out of the fits area
        values[(x == -1) | (y == -1)] = default
        values[np.isnan(values)] = default

        if pad_index:
            return np.array(["{0}_{1}".format(root, str(int(val)).zfill(int(np.ceil(np.log10(len(set(values))+1))))) for val in values])
        else:
            return np.array(["{0}_{1}".format(root, int(val)) for val in values])
